<?php

/**
 * JCH Optimize - Joomla! plugin to aggregate and minify external resources for
 * optmized downloads
 * @author Samuel Marshall <sdmarshall73@gmail.com>
 * @copyright Copyright (c) 2010 Samuel Marshall
 * @license GNU/GPLv3, See LICENSE file
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * If LICENSE file missing, see <http://www.gnu.org/licenses/>.
 */
defined('_JCH_EXEC') or die('Restricted access');

/**
 * Class to parse HTML and find css and js links to replace, populating an array with matches
 * and removing found links from HTML
 * 
 */
class JchOptimizeParser extends JchOptimizeBase
{

        /** @var string   Html of page */
        public $sHtml = '';

        /** @var array    Array of css or js urls taken from head */
        protected $aLinks = array();
        protected $aUrls = array();
        public $params    = null;
        public $sLnEnd    = '';
        public $sTab      = '';
        public $sFileHash = '';
        private $bPreserveOrder;
        
        protected $oFileRetriever;

        /**
         * Constructor
         * 
         * @param JRegistry object $params      Plugin parameters
         * @param string  $sHtml                Page HMTL
         */
        public function __construct($oParams, $sHtml, $oFileRetriever)
        {
                $this->params = $oParams;
                $this->sHtml  = $sHtml;
                
                $this->oFileRetriever = $oFileRetriever;

                $this->sLnEnd = JchPlatformUtility::lnEnd();
                $this->sTab   = JchPlatformUtility::tab();

                if (!defined('JCH_TEST_MODE'))
                {
                        $this->sFileHash = serialize($this->params) . JCH_VERSION;
                }

                $this->parseHtml();
        }

        /**
         * 
         * @return type
         */
        public function cleanHtml()
        {
                return preg_replace(array($this->getHeadRegex(),
                        implode('', $this->getJsRegex()), implode('', $this->getCssRegex())), '', $this->sHtml);
        }

        /**
         * 
         */
        public function getHtmlHash()
        {
                $sHtmlHash = '';

                preg_replace_callback('#<(?!/)[^>]++>#i',
                                      function($aM) use (&$sHtmlHash)
                {
                        $sHtmlHash .= $aM[0];

                        return;
                }, $this->cleanHtml(), 200);

                return $sHtmlHash;
        }

        /**
         * Removes applicable js and css links from search area
         * 
         */
        public function parseHtml()
        {
                $oParams = $this->params;

                $aCBArgs = array();

                $this->getHeadHtml();
                $this->getBodyHtml();

                $aExJsComp  = $this->getExComp($oParams->get('excludeJsComponents', ''));
                $aExCssComp = $this->getExComp($oParams->get('excludeCssComponents', ''));

                $aExcludeJs  = JchOptimize::getArray($oParams->get('excludeJs', ''));
                $aExcludeCss = JchOptimize::getArray($oParams->get('excludeCss', ''));

                $aCBArgs['excludes']['js']     = array_merge($aExcludeJs, $aExJsComp);
                $aCBArgs['excludes']['css']    = array_merge($aExcludeCss, $aExCssComp);
                $aCBArgs['excludes']['script'] = JchOptimize::getArray($oParams->get('pro_excludeScripts'));

                if ($oParams->get('javascript', 1))
                {
                        $this->initSearch('js', $aCBArgs);
                }

                if ($this->enableCssCompression())
                {
                        $this->initSearch('css', $aCBArgs);
                }
        }

        /**
         * 
         * @param type $sType
         */
        protected function initSearch($sType, $aCBArgs)
        {
                $aCBArgs['type'] = $sType;

                $aRegex = $this->{'get' . ucfirst($sType) . 'Regex'}();
                $sRegex = implode('', $aRegex);

                $this->{'iIndex_' . $sType} = -1;
                $this->bExclude             = TRUE;

                $this->searchArea($sRegex, $sType, 'head', $aCBArgs);

                if ($this->params->get('pro_searchBody', '0'))
                {
                        $this->searchArea($sRegex, $sType, 'body', $aCBArgs);
                }
        }

        /**
         * 
         * @param type $sRegex
         * @param type $sType
         * @param type $sSection
         * @param type $aCBArgs
         * @throws Exception
         */
        protected function searchArea($sRegex, $sType, $sSection, $aCBArgs)
        {
                JCH_DEBUG ? JchPlatformProfiler::mark('beforeSearchArea' . $sType . ' - ' . $sSection . ' plgSystem (JCH Optimize)') : null;

                $obj = $this;

                $this->bPreserveOrder = (bool) !(($sType == 'css' && $this->params->get('pro_optimizeCssDelivery', '0'))
                        || ($this->params->get('bottom_js', '0'))
                        || ($sType == 'js' && $this->params->get('bottom_js', '0') == '1'));

                $sProcessedHtml = preg_replace_callback($sRegex,
                                                        function($aMatches) use ($obj, $aCBArgs, $sSection)
                {
                        return $obj->replaceScripts($aMatches, $aCBArgs, $sSection);
                }, $this->{'s' . ucfirst($sSection) . 'Html'});

                JCH_DEBUG ? JchPlatformProfiler::mark('afterSearchArea' . $sType . ' - ' . $sSection . ' plgSystem (JCH Optimize)') : null;

                if (is_null($sProcessedHtml))
                {
                        throw new Exception(sprintf(JchPlatformUtility::translate('Error while parsing for %1$s links in %2$s' .
                                        ' ...turn off combine %1$s option'), $sType, $sSection));
                }

                $this->{'s' . ucfirst($sSection) . 'Html'} = $sProcessedHtml;
        }

        /**
         * 
         * @return type
         */
        public function enableCssCompression()
        {
                if (JchPlatformUtility::isMsieLT10())
                {
                        return FALSE;
                }
                else
                {
                        return ($this->params->get('css', 1) || $this->params->get('csg_enable', 0));
                }
        }

        /**
         * Callback function used to remove urls of css and js files in head tags
         *
         * @param array $aMatches       Array of all matches
         * @return string               Returns the url if excluded, empty string otherwise
         */
        public function replaceScripts($aMatches, $aCBArgs, $sSection)
        {
                if ((strpos($aMatches[0], '<!--') === 0)
                        || (!isset($aMatches[1]) || trim($aMatches[1]) == '' || trim($aMatches[1]) == '/')
                        && (!isset($aMatches[3]) || trim($aMatches[3]) == ''))
                {
                        return $aMatches[0];
                }

                $sType = $aCBArgs['type'];

                $sUrl         = $aMatches[1];
                $sFile        = isset($aMatches[2]) ? $aMatches[2] : '';
                $sDeclaration = isset($aMatches[3]) ? $aMatches[3] : '';
                $aExcludes    = array();
                $sPath        = '';

                if (isset($aCBArgs['excludes']))
                {
                        $aExcludes = $aCBArgs['excludes'];
                }

                $aExcludes['script'] = array_map(function($sScript)
                {
                        return stripslashes($sScript);
                }, $aExcludes['script']);

                $aExcludes['js']  = array_merge($aExcludes['js'], array('.com/maps/api/js', '.com/jsapi', '.com/uds', 'typekit.net'),
                                                JchPlatformExcludes::head('js'));
                $aExcludes['css'] = array_merge($aExcludes['css'], array('fonts.googleapis.com'), JchPlatformExcludes::head('css'));

                if ($sSection == 'body')
                {
                        $aExcludes['script'] = array_merge($aExcludes['script'], array('document.write'), JchPlatformExcludes::body('js', 'script'));
                        $aExcludes['js']     = array_merge($aExcludes['js'], array('.com/recaptcha/api'), JchPlatformExcludes::body('js'));
                }

                if ($sUrl != '')
                {
                        $sPath .= JchOptimizeHelper::getFilePath($sUrl);
                }

                $sMedia = '';

                if (($sType == 'css') && (preg_match('#media=(?(?=["\'])(?:["\']([^"\']+))|(\w+))#i', $aMatches[0], $aMediaTypes) > 0))
                {
                        $sMedia .= $aMediaTypes[1] ? $aMediaTypes[1] : $aMediaTypes[2];
                }
                //JCH_DEBUG ? JchPlatformProfiler::mark('beforeReplaceScript - ' . $sFile . ' plgSystem (JCH Optimize)') : null;

                switch (TRUE)
                {
                        case (($sUrl != '') && $this->isDuplicated($sUrl)):
                                
                                $this->bExclude = FALSE;
                                
                                return '';
                                
                        case (($sUrl != '') && !empty($aExcludes[$sType]) && JchOptimizeHelper::findExcludes($aExcludes[$sType], $sUrl)):
                        case (($sPath != '') && $this->isHttpAdapterAvailable($sPath)):
                        case ($sUrl != '' && preg_match('#^https#', $sUrl) && !extension_loaded('openssl')):
                        case ($sUrl != '' && preg_match('#^data:#', $sUrl)):
                        case ($sDeclaration != '' && $this->excludeDeclaration($sType)):
                        case ($sDeclaration != '' && JchOptimizeHelper::findExcludes($aExcludes['script'], $sDeclaration, TRUE)):
                        case (($sUrl != '') && $this->excludeExternalExtensions($sUrl)):

                                //JCH_DEBUG ? JchPlatformProfiler::mark('afterReplaceScript - ' . $sFile . ' plgSystem (JCH Optimize)') : null;

                                $this->bExclude = TRUE;

                                return $aMatches[0];

                        default:
                                $return = '';

                                if ($this->bExclude && $this->bPreserveOrder)
                                {
                                        $this->bExclude = FALSE;

                                        $iIndex = ++$this->{'iIndex_' . $sType};
                                        $return = '<JCH_' . strtoupper($sType) . $iIndex . '>';
                                }
                                elseif (!$this->bPreserveOrder)
                                {
                                        $iIndex = 0;
                                }
                                else
                                {
                                        $iIndex = $this->{'iIndex_' . $sType};
                                }

                                $array = array();

                                $array['match'] = $aMatches[0];

                                if ($sUrl == '' && trim($sDeclaration) != '')
                                {
                                        $content = str_replace(array('<!--', '-->'), '', $sDeclaration);

                                        $array['content'] = $content;
                                        $id               = $content;
                                }
                                else
                                {
                                        $array['url']  = $sUrl;
                                        $array['path'] = $sPath;
                                        $array['file'] = $sFile;
                                        $id            = $sUrl;
                                }

                                if ($this->sFileHash != '')
                                {
                                        $array['id'] = md5($this->sFileHash . $id);
                                }

                                if ($sType == 'css')
                                {
                                        $array['media'] = $sMedia;
                                }

                                $this->aLinks[$sType][$iIndex][] = $array;

                                return $return;
                }
        }

        /**
         * 
         * @param type $sUrl
         */
        protected function isDuplicated($sUrl)
        {
                $return = FALSE;
                $sUrl = preg_replace('#https?:#i', '', $sUrl);
                
                if($this->params->get('remove_duplicates', '1'))
                {
                        $return = in_array($sUrl, $this->aUrls);
                }
                
                if(!$return)
                {
                        $this->aUrls[] = $sUrl;
                }
                
                return $return;
        }
        /**
         * 
         * @param type $sPath
         */
        protected function excludeExternalExtensions($sPath)
        {
                if ($this->params->get('excludeAllExtensions', '1'))
                {
                        return !JchOptimizeHelper::isInternal($sPath) || preg_match('#' . JchPlatformExcludes::extensions() . '#i', $sPath);
                }
                
                return FALSE;
        }
        
        /**
         * Generates regex for excluding components set in plugin params
         * 
         * @param string $param
         * @return string
         */
        protected function getExComp($sExComParam)
        {
                $aComponents = array_filter(JchOptimize::getArray($sExComParam));
                $aExComp     = array();

                if (!empty($aComponents))
                {
                        $aExComp = array_map(function($sValue)
                        {
                                return $sValue . '/';
                        }, $aComponents);
                }

                return $aExComp;
        }

        /**
         * Fetches Class property containing array of matches of urls to be removed from HTML
         * 
         * @return array
         */
        public function getReplacedFiles()
        {
                return $this->aLinks;
        }

        /**
         * Set the Searcharea property
         * 
         * @param type $sSearchArea
         */
        public function setSearchArea($sSearchArea, $sSection)
        {
                $this->{'s' . ucfirst($sSection) . 'Html'} = $sSearchArea;
        }

        /**
         * Determines if document is of html5 doctype
         * 
         * @return boolean
         */
        public function isHtml5()
        {
                if (preg_match('#^<!DOCTYPE html>#i', trim($this->sHtml)))
                {
                        return true;
                }
                else
                {
                        return false;
                }
        }

        /**
         * 
         * @return string
         */
        protected static function ifRegex()
        {
                return '<!--(?>-?[^-]*+)*?-->';
        }

        
}
